#!/usr/bin/env python
# -*- coding: utf-8 -*-

from rdflib import *
from rdflib.util import guess_format
import rdflib
import csv
import json
import sys, collections
import requests
import pandas
import re
import os
from six import text_type as unicode

from jinja2 import Template
from toposort import toposort_flatten
from numpy import isnan
import uuid
import tempfile
import ijson
from . import iterparse_filter
#import xml.etree.ElementTree as ET
import xml.etree.ElementTree

from itertools import chain

import zipfile
import gzip

import logging

import hashlib
from slugify import slugify

def hash(value):
    m = hashlib.sha256()
    m.update(value.encode('utf-8'))
    return m.hexdigest()

csvw = Namespace('http://www.w3.org/ns/csvw#')
ov = Namespace('http://open.vocab.org/terms/')
setl = Namespace('http://purl.org/twc/vocab/setl/')
prov = Namespace('http://www.w3.org/ns/prov#')
pv = Namespace('http://purl.org/net/provenance/ns#')
sp = Namespace('http://spinrdf.org/sp#')
sd = Namespace('http://www.w3.org/ns/sparql-service-description#')
dc = Namespace('http://purl.org/dc/terms/')
void = Namespace('http://rdfs.org/ns/void#')
api_vocab = Namespace('http://purl.org/linked-data/api/vocab#')

sys.setrecursionlimit(10000)

from requests_testadapter import Resp

def camelcase(s):
    return slugify(s).title().replace("-","")

class LocalFileAdapter(requests.adapters.HTTPAdapter):
    def build_response_from_file(self, request):
        file_path = request.url[7:]
        with open(file_path, 'rb') as file:
            buff = bytearray(os.path.getsize(file_path))
            file.readinto(buff)
            resp = Resp(buff)
            r = self.build_response(request, resp)
            return r
    def send(self, request, stream=False, timeout=None,
             verify=True, cert=None, proxies=None):
        return self.build_response_from_file(request)

requests_session = requests.session()
requests_session.mount('file://', LocalFileAdapter())
requests_session.mount('file:///', LocalFileAdapter())
    
datatypeConverters = collections.defaultdict(lambda: str)
datatypeConverters.update({
    XSD.string: str,
    XSD.decimal: float,
    XSD.integer: int,
    XSD.float: float,
    XSD.double: float
})

run_samples = False

_rdf_formats_to_guess = [
    'xml',
    'json-ld',
    'trig',
    'nquads',
    'trix'
]

def lru(original_function, maxsize=1000):
    mapping = {}

    PREV, NEXT, KEY, VALUE = 0, 1, 2, 3         # link fields
    head = [None, None, None, None]        # oldest
    tail = [head, None, None, None]   # newest
    head[NEXT] = tail

    def fn(*args, **kw):
        key = (args,tuple(kw.items()))
        PREV, NEXT = 0, 1
        #print "Cache lookup for "+str(key)
        link = mapping.get(key, head)
        if link is head:
            #print "Cache miss for "+str(key)
            value = original_function(*args,**kw)
            if len(mapping) >= maxsize:
                old_prev, old_next, old_key, old_value = head[NEXT]
                head[NEXT] = old_next
                old_next[PREV] = head
                del mapping[old_key]
            last = tail[PREV]
            link = [last, tail, key, value]
            mapping[key] = last[NEXT] = tail[PREV] = link
        else:
            #print "Cache hit for "+str(key)
            link_prev, link_next, key, value = link
            link_prev[NEXT] = link_next
            link_next[PREV] = link_prev
            last = tail[PREV]
            last[NEXT] = tail[PREV] = link
            link[PREV] = last
            link[NEXT] = tail
        return value
    return fn

def read_csv(location, result):
    args = dict(
        sep = result.value(csvw.delimiter, default=Literal(",")).value,
        #header = result.value(csvw.headerRow, default=Literal(0)).value),
        skiprows = result.value(csvw.skipRows, default=Literal(0)).value,
        dtype = object
    )
    if result.value(csvw.header):
        args['header'] = [0]
    df = pandas.read_csv(get_content(location, result),encoding='utf-8', **args)
    logger.debug("Loaded %s", location)
    return df
        
def read_graph(location, result, g = None):
    if g is None:
        g = ConjunctiveGraph()
    graph = ConjunctiveGraph(store=g.store, identifier=result.identifier)
    if len(graph) == 0:
        data = get_content(location, result).read()
        f = guess_format(location)
        for fmt in [f] + _rdf_formats_to_guess:
            try:
                graph.parse(data=data, format=fmt)
                break
            except Exception as e:
                #print e
                pass
        if len(graph) == 0:
            logger.error("Could not parse graph: %s", location)
        if result[RDF.type:OWL.Ontology]:
            for ontology in graph.subjects(RDF.type, OWL.Ontology):
                imports = [graph.resource(x) for x in graph.objects(ontology, OWL.imports)]
                for i in imports:
                    read_graph(i.identifier, i, g = g)
    return g

class FileLikeFromIter(object):
    def __init__(self, content_iter):
        self.iter = content_iter
        self.data = b''

    def __iter__(self):
        return self.iter

    def read(self, n=None):
        if n is None:
            return self.data + ''.join(l for l in self.iter)
        else:
            while len(self.data) < n:
                try:
                    self.data = b''.join((self.data, next(self.iter)))
                except StopIteration:
                    break
            result, self.data = self.data[:n], self.data[n:]
            return result

def _open_local_file(location):
    if location.startswith("file://"):
        if os.name == 'nt': # skip the initial 
            return open(location.replace('file:///','').replace('file://',''),'rb')
        else:
            return open(location.replace('file://',''),'rb')

content_handlers = [
    _open_local_file,
    lambda location: FileLikeFromIter(requests.get(location,stream=True).iter_content(1024*1024))
]
        
def get_content(location, result):
    response = None
    for handler in content_handlers:
        response = handler(location)
        if response is not None:
            break
    if result[RDF.type:setl.Tempfile]:
        result = to_tempfile(response)
        
    for t in result[RDF.type]:
        # Do we know how to unpack this?
        if t.identifier in unpackers:
            response = unpackers[t.identifier](response)
    return response

def to_tempfile(f):
    tf = tempfile.TemporaryFile()
    logger.debug("Writing %s to disk.", f)
    for chunk in f:
        if chunk: # filter out keep-alive new chunks
            tf.write(chunk)
    tf.seek(0)
    logger.debug("Finished writing %s to disk.", f)
    return tf

def unpack_zipfile(f):
    zf = zipfile.ZipFile(f, mode='r')
    files = zf.infolist()
    return zf.open(files[0])

unpackers = {
#    setl.Tempfile : lambda x: x,
    setl.ZipFile : lambda x: unpack_zipfile(to_tempfile(x)),
    setl.GZipFile : lambda f: gzip.GzipFile(fileobj=f,mode='r')
}

def read_excel(location, result):
    args = dict(
        sheetname = result.value(setl.sheetname, default=Literal(0)).value,
        header = [int(x) for x in result.value(csvw.headerRow, default=Literal('0')).value.split(',')],
        skiprows = result.value(csvw.skipRows, default=Literal(0)).value
    )
    if result.value(csvw.header):
        args['header'] = [result.value(csvw.header).value]
    df = pandas.read_excel(get_content(location, result),encoding='utf-8', **args)
    return df

def read_xml(location, result):
    validate_dtd = False
    if result[RDF.type:setl.DTDValidatedXML]:
        validate_dtd = True
    f = iterparse_filter.IterParseFilter(validate_dtd=validate_dtd)
    if result.value(setl.xpath) is None:
        logger.debug("no xpath to select on from %s", location)
        f.iter_end("/*")
    for xp in result[setl.xpath]:
        f.iter_end(xp.value)
    for (i, (event, ele)) in enumerate(f.iterparse(get_content(location, result))):
        yield i, ele

             
def read_json(location, result):
    selector = result.value(api_vocab.selector)
    if selector is not None:
        selector = selector.value
    else:
        selector = ""
    return enumerate(ijson.items(get_content(location, result), selector))
            

extractors = {
    setl.XPORT : lambda location, result: pandas.read_sas(get_content(location, result), format='xport'),
    setl.SAS7BDAT : lambda location, result: pandas.read_sas(get_content(location, result), format='sas7bdat'),
    setl.Excel : read_excel,
    csvw.Table : read_csv,
    OWL.Ontology : read_graph,
    void.Dataset : read_graph,
    setl.JSON : read_json,
    setl.XML : read_xml,
    URIRef("https://www.iana.org/assignments/media-types/text/plain") : lambda location, result: get_content(location, result)
}

    
try:
    from bs4 import BeautifulSoup
    extractors[setl.HTML] = lambda location, result: BeautifulSoup(get_content(location, result).read(), 'html.parser')
except Exception as e:
    pass
    
    
def load_csv(csv_resource):
    column_descriptions = {}
    for col in csv_resource[csvw.column]:
        label = col.value(RDFS.label).value
        column_descriptions[label] = col
    csv_graph = Graph(identifier=csv_resource)
    s = [x for x in csv.reader(open(str(csv_resource.value(csvw.url).identifier).replace("file://","")),
                   delimiter=str(csv_resource.value(csvw.delimiter,default=",").value),
                   quotechar=str(csv_resource.value(csvw.quoteChar,default='"').value))]
    header = None
    properties = []
    propertyMap = {}
    skip_value = csv_resource.value(csvw.null)
    if skip_value is not None:
        skip_value = skip_value.value
    for i, r in enumerate(s):
        if header is None:
            header = r
            for j, h in enumerate(header):
                col_desc = None
                if h in column_descriptions:
                    col_desc = column_descriptions[h]
                col = csv_graph.resource(URIRef("urn:col_"+str(h)))
                col.add(RDFS.label, Literal(h))
                col.add(ov.csvCol, Literal(j))
                if col_desc is not None:
                    col.add(RDFS.range, col_desc.value(RDFS.range, default=XSD.string))
                properties.append(col)
                propertyMap[h] = col
            continue
        res = csv_graph.resource(csv_resource.identifier+"_row_"+str(i))
        res.add(RDF.type, csvw.Row)
        res.add(csvw.rownum, Literal(i))
        for j, value in enumerate(r):
            if skip_value is not None and skip_value == value:
                continue
            #print i, j, value
            prop = properties[j]
            datatype = prop.value(RDFS['range'], default=XSD.string)
            lit =  Literal(value, datatype=datatype.identifier)
            #print i, prop.identifier, lit.n3()
            res.add(prop.identifier, lit)
    logger.debug("Table has %s rows, %s columns, and %s triples", len(s), len(header), len(csv_graph))
    return csv_graph

formats = {
    None:'xml',
    "application/rdf+xml":'xml',
    "text/rdf":'xml',
    'text/turtle':'turtle',
    'application/turtle':'turtle',
    'application/x-turtle':'turtle',
    'text/plain':'nt',
    'text/n3':'n3',
    'application/trig':'trig',
    'application/json':'json-ld'
}

def create_python_function(f, resources):
    global_vars = {'this' : f, 'resources': resources}
    local_vars = {}
    script = f.value(prov.value)
    for qd in f[prov.qualifiedDerivation]:
        entity = resources[qd.value(prov.entity).identifier]
        name = qd.value(prov.hadRole).value(dc.identifier)
        local_vars[name.value] = entity
    exec(script.value, local_vars, global_vars)
    resources[f.identifier] = global_vars['result']
        
def get_order(setl_graph):
    nodes = collections.defaultdict(set)

    for typ in actions:
        for task in setl_graph.subjects(RDF.type, typ):
            task = setl_graph.resource(task)
            for used in task[prov.used]:
                nodes[task.identifier].add(used.identifier)

            for usage in task[prov.qualifiedUsage]:
                used = usage.value(prov.entity)
                nodes[task.identifier].add(used.identifier)
            for generated in task.subjects(prov.wasGeneratedBy):
                nodes[generated.identifier].add(task.identifier)
            for derivation in task[prov.qualifiedDerivation]:
                derived = derivation.value(prov.entity)
                nodes[task.identifier].add(derived.identifier)
    
    return toposort_flatten(nodes)

def extract(e, resources):
    logger.info('Extracting %s',e.identifier)
    used = e.value(prov.used)
    for result in e.subjects(prov.wasGeneratedBy):
        if used is None:
            used = result
        for t in result[RDF.type]:
            # Do we know how to generate this?
            if t.identifier in extractors:
                logger.info("Extracted %s", used.identifier)
                resources[result.identifier] = extractors[t.identifier](used.identifier, result)
                return resources[result.identifier]

def isempty(value):
    try:
        return isnan(value)
    except:
        return value is None

def clone(value):
    __doc__ = '''This is only a JSON-level cloning of objects. Atomic objects are invariant, and don't need to be cloned.'''
    if isinstance(value, list):
        return [x for x in value]
    elif isinstance(value, dict):
        return dict(value)
    else:
        return value

functions = {}
def get_function(expr, local_keys):
    key = tuple([expr]+sorted(local_keys))
    if key not in functions:
        script = '''lambda %s: %s'''% (', '.join(sorted(local_keys)), expr)
        fn = eval(script)
        fn.__name__ = str(expr)
        functions[key] = fn
    return functions[key]

templates = {}
def get_template(templ):
    if templ not in templates:
        t = Template(templ)
        templates[templ] = t
    return templates[templ]
    
def process_row(row, template, rowname, table, resources, transform, variables):
    result = []
    e = {'row':row,
         'name': rowname,
         'table': table,
         'resources': resources,
         'template': template,
         "transform": transform,
         "setl_graph": transform.graph,
         "isempty":isempty,
         "slugify" : slugify,
         "camelcase" : camelcase,
         "hash":hash,
         "isinstance":isinstance,
         "str":str,
         "float":float,
         "int":int,
         "chain": lambda x: chain(*x),
         "list":list
    }
    e.update(variables)
    e.update(rdflib.__dict__)
    todo = [[x, result, e] for x in template]

    while len(todo) > 0:
        task, parent, env = todo.pop()
        key = None
        value = task
        this = None
        if isinstance(parent, dict):
            if len(task) != 2:
                logger.debug(task)
            key, value = task
            kt = get_template(key)
            key = kt.render(**env)
        if isinstance(value, dict):
            if '@if' in value:
                try:
                    fn = get_function(value['@if'], env.keys())
                    incl = fn(**env)
                    if incl is None or not incl:
                        continue
                except KeyError:
                    continue
                except AttributeError:
                    continue
                except TypeError:
                    continue
                except Exception as e:
                    trace = sys.exc_info()[2]
                    logger.error("Error in conditional %s\nRelevant Environment:", value['@if'])
                    for key, v in env.items():
                        #if key in value['@if']:
                        if hasattr(v, 'findall'):
                            v = xml.etree.ElementTree.tostring(v)
                        logger.error(key + "\t" + str(v)[:1000])
                    raise(e, None, trace)
            if '@for' in value:
                f = value['@for']
                if isinstance(f, list):
                    f = ' '.join(f)
                variable_list, expression = f.split(" in ", 1)
                variable_list = re.split(',\s+', variable_list.strip())
                val = value
                if '@do' in value:
                    val = value['@do']
                else:
                    del val['@for']
                try:
                    fn = get_function(expression, env.keys())
                    values = fn(**env)
                    if values is not None:
                        for v in values:
                            if len(variable_list) == 1:
                                v = [v]
                            new_env = dict(env)
                            for i, variable in enumerate(variable_list):
                                new_env[variable] = v[i]
                            child = clone(val)
                            todo.append((child, parent, new_env))
                except KeyError:
                    pass
                except Exception as e:
                    trace = sys.exc_info()[2]
                    logger.error("Error in @for: %s", value['@for'])
                    logger.error("Locals: %s", env.keys())
                    raise(e, None, trace)
                continue
            if '@with' in value:
                f = value['@with']
                if isinstance(f, list):
                    f = ' '.join(f)
                expression, variable_list = f.split(" as ", 1)
                variable_list = re.split(',\s+', variable_list.strip())
                val = value
                if '@do' in value:
                    val = value['@do']
                else:
                    del val['@with']
                try:
                    fn = get_function(expression, env.keys())
                    v = fn(**env)
                    if v is not None:
                        if len(variable_list) == 1:
                            v = [v]
                        new_env = dict(env)
                        for i, variable in enumerate(variable_list):
                            new_env[variable] = v[i]
                        child = clone(val)
                        todo.append((child, parent, new_env))
                except KeyError:
                    pass
                except Exception as e:
                    trace = sys.exc_info()[2]
                    logger.error("Error in with: %s", value['@with'])
                    logger.error("Locals: %s", env.keys())
                    raise(e, None, trace)
                continue
            this = {}
            for child in value.items():
                if child[0] == '@if':
                    continue
                if child[0] == '@for':
                    continue
                todo.append((child, this, env))
        elif isinstance(value, list):
            this = []
            for child in value:
                todo.append((child, this, env))
        elif isinstance(value, unicode):
            try:
                template = get_template(unicode(value))
                this = template.render(**env)
            except Exception as e:
                trace = sys.exc_info()[2]
                logger.error("Error in template %s %s", value, type(value))
                logger.error("Relevant Environment:")
                for key, v in env.items():
                    #if key in value:
                    if hasattr(v, 'findall'):
                        v = xml.etree.ElementTree.tostring(v)
                    logger.error(key + "\t" + str(v)[:1000])
                raise(e, None, trace)
        else:
            this = value

        if key is not None:
            parent[key] = this
        else:
            parent.append(this)
    return result

def json_transform(transform, resources):
    logger.info("Transforming %s", transform.identifier)
    tables = [u for u in transform[prov.used]]
    variables = {}
    for usage in transform[prov.qualifiedUsage]:
        used = usage.value(prov.entity)
        role = usage.value(prov.hadRole)
        roleID  = role.value(dc.identifier)
        variables[roleID.value] = resources[used.identifier]
        #print "Using", used.identifier, "as", roleID.value
    
    generated = list(transform.subjects(prov.wasGeneratedBy))[0]
    logger.info("Generating %s", generated.identifier)

    if generated.identifier in resources:
        result = resources[generated.identifier]
    else:
        result = ConjunctiveGraph()
        if generated[RDF.type : setl.Persisted]:
            result = ConjunctiveGraph(store="Sleepycat")
        if generated[RDF.type : setl.Persisted]:
            _, tempdir = tempfile.mkstemp()
            logger.info("Persisting %s to %s", generated.identifier, tempdir)
            result.store.open(tempdir, True)
    s = transform.value(prov.value).value
    try:
        jslt = json.loads(s)
    except Exception as e:
        trace = sys.exc_info()[2]
        if 'No JSON object could be decoded' in e.message:
            logger.error(s)
        if 'line' in e.message:
            line = int(re.search("line ([0-9]+)", e.message).group(1))
            logger.error("Error in parsing JSON Template at line %d:", line)
            logger.error('\n'.join(["%d: %s"%(i+line-3, x) for i, x in enumerate(s.split("\n")[line-3:line+4])]))
        raise(e, None, trace)
    context = transform.value(setl.hasContext)
    if context is not None:
        context = json.loads(context.value)
    for t in tables:
        logger.info("Using %s", t.identifier)
        table = resources[t.identifier]
        it = table
        if isinstance(table, pandas.DataFrame):
            #if run_samples:
            #    table = table.head()
            it = table.iterrows()
            logger.info("Transforming %s rows.", len(table.index))
        else:
            logger.info("Transforming %s", t.identifier)
        for rowname, row in it:
            if run_samples and rowname >= 100:
                break
            try:
                root = {
                    "@id": generated.identifier,
                    "@graph": process_row(row, jslt, rowname, table, resources, transform, variables)
                }
                if context is not None:
                    root['@context'] = context
                before = len(result)
                #graph = ConjunctiveGraph(identifier=generated.identifier)
                #graph.parse(data=json.dumps(root),format="json-ld")
                data = json.dumps(root)
                del root
                result.parse(data=data, format="json-ld")
                del data
                after = len(result)
                logger.debug("Row "+str(rowname)+" added "+str(after-before)+" triples.")
                sys.stdout.flush()
            except Exception as e:
                trace = sys.exc_info()[2]
                if isinstance(table, pandas.DataFrame):
                    logger.error("Error on %s %s", rowname, row)
                else:
                    logger.error("Error on %s", rowname)
                raise(e, None, trace)
                
    resources[generated.identifier] = result

def transform(transform_resource, resources):
    logger.info('Transforming %s',transform_resource.identifier)

    transform_graph = ConjunctiveGraph()
    for result in transform_graph.subjects(prov.wasGeneratedBy):
        transform_graph = ConjunctiveGraph(identifier=result.identifier)

    used = set(transform_resource[prov.used])
    
    for csv in [u for u in used if u[RDF.type:csvw.Table]]:
        csv_graph = Graph(store=transform_graph.store, identifier=csv)
        csv_graph += graphs[csv.identifier]

    
    for script in [u for u in used if u[RDF.type:setl.PythonScript]]:
        logger.info("Script: %s", script.identifier)
        s = script.value(prov.value).value
        l = dict(graph = transform_graph, setl_graph = transform_resource.graph)
        gl = dict()
        exec(s, gl, l)

    for jsldt in [u for u in used if u[RDF.type:setl.PythonScript]]:
        logger.info("Script: %s", script.identifier)
        s = script.value(prov.value).value
        l = dict(graph = transform_graph, setl_graph = transform_resource.graph)
        gl = dict()
        exec(s, gl, l)

    for update in [u for u in used if u[RDF.type:sp.Update]]:
        logger.info("Update: %s", update.identifier)
        query = update.value(prov.value).value
        transform_graph.update(query)
        
    for construct in [u for u in used if u[RDF.type:sp.Construct]]:
        logger.info("Construct: %s", construct.identifier)
        query = construct.value(prov.value).value
        g = transform_graph.query(query)
        transform_graph += g
        
    for csv in [u for u in used if u[RDF.type:csvw.Table]]:
        g = Graph(identifier=csv.identifier,store=transform_graph.store)
        g.remove((None, None, None))
        transform_graph.store.remove_graph(csv.identifier)
            
    for result in transform_graph.subjects(prov.wasGeneratedBy):
        graphs[result.identifier] = transform_graph

        
def load(load_resource, resources):
    logger.info('Loading %s',load_resource.identifier)
    file_graph = Dataset(default_union=True)
    to_disk = False
    for used in load_resource[prov.used]:
        if used[RDF.type : setl.Persisted]:
            to_disk = True
            file_graph = Dataset(store='Sleeptcat', default_union=True)
            _, tempdir = tempfile.mkstemp()
            logger.debug("Gathering %s into %s", load_resource.identifier, tempdir)
            file_graph.store.open(tempdir, True)
            break
    if len(list(load_resource[prov.used])) == 1:
        logger.info("Using %s",load_resource.value(prov.used).identifier)
        file_graph = resources[load_resource.value(prov.used).identifier]
    else:
        for used in load_resource[prov.used]:
            logger.info("Using %s",used.identifier)
            used_graph = resources[used.identifier]
            file_graph.namespace_manager = used_graph.namespace_manager
            #print used_graph.serialize(format="trig")
            file_graph.addN(used_graph.quads())

    for generated in load_resource.subjects(prov.wasGeneratedBy):
        # TODO: support LDP-based loading
        if generated[RDF.type:pv.File]:
            fmt = generated.value(dc['format'])
            if fmt is not None:
                fmt = fmt.value
            if fmt in formats:
                fmt = formats[fmt]
                #print fmt
            with open(generated.identifier.replace("file://",''), 'wb') as o:
                o.write(file_graph.serialize(format=fmt))
                o.close()
        elif generated[RDF.type:sd.Service]:
            from rdflib.plugins.stores.sparqlstore import SPARQLUpdateStore
            endpoint = generated.value(sd.endpoint, default=generated).identifier
            store = SPARQLUpdateStore(endpoint, endpoint, autocommit=False)
            endpoint_graph = Dataset(store=store, identifier=generated.identifier, default_union=True)
            endpoint_graph.addN(file_graph.quads())
            endpoint_graph.commit()
    #if to_disk:
    #    file_graph.close()
    
        
actions = {
    setl.Extract : extract,
    setl.Transform : json_transform,
    setl.Load : load,
    setl.PythonScript : create_python_function,
    setl.IsEmpty : isempty
}
            
def _setl(setl_graph):
    global logger
    if logger is None:
        logger = logging.getLogger(__name__)
    resources = {}
    resources.update(actions)

    tasks = [setl_graph.resource(t) for t in get_order(setl_graph)]

    for task in tasks:
        action = [actions[t.identifier] for t in task[RDF.type] if t.identifier in actions]
        if len(action) > 0:
            action[0](task, resources)
    return resources
logger = None

def main():
    args = sys.argv[1:]
    logging_level = logging.DEBUG
    if '-q' in args or '--quiet' in args:
        logging_level = logging.WARNING
    logging.basicConfig(level=logging_level)

    global logger
    logger = logging.getLogger(__name__)
    
    global run_samples
    setl_file = args[0]
    if 'sample' in args:
        run_samples = True
        logger.warning("Only processing a few sample rows.")
    setl_graph = ConjunctiveGraph()
    content = open(setl_file).read()
    setl_graph.parse(data=content, format="turtle")

    graphs = _setl(setl_graph)
#    print "Finished processing"
#    return graphs

if __name__ == '__main__':
    result = main()
    logger.info("Exiting")



"""An experimental XPath-based streaming filter for ElementTree's iterparse

For details see:
http://dalkescientific.com/writings/diary/archive/2006/11/06/iterparse_filter.html
"""
# I have got to rearrange my site to use shorter URLs.

__version__ = "0.9-experimental"

import re

dtd_validation = False
try:
  from lxml import etree
  dtd_validation = True
except ImportError:
  try:
    # Python 2.5
    import xml.etree.cElementTree as etree
  except ImportError:
    try:
      # Python 2.5
      import xml.etree.ElementTree as etree
    except ImportError:
      try:
        # normal cElementTree install
        import cElementTree as etree
      except ImportError:
          # normal ElementTree install
          import elementtree.ElementTree as etree

# define "letter" as "any character except /:[]()@={}* or in \s"
#   (XXX make it match the XML spec)
# A URI is:
#   letter+
#   letter+ ':' letter+   --- a namespace prefixed term, like xml:space
#   '{'  [^}]*  '}' letter+  --- a Clark namespace term, like {http://a}b
# Can also use a '*' in place of a URI or in the tag part of a namespaced field
#
# URIs are separated only be '/' and '//'.
# These may not occur together, eg, '///' is not allowed.

# Basing this tokenization method in part on elementtree.ElementPath
xpath_tokenizer = re.compile( r"""
(// | / )  # separators

| (?:  # namespaced term
  ([^\/\:\[\]\(\)\@\=\{\}\*\s]+) :    # namespace
  ([^\/\:\[\]\(\)\@\=\{\}\*\s]+|\*)      # tag
) 

| (?:
  \{([^}]*)\}      # namespace in Clark notation
  ([^\/\:\[\]\(\)\@\=\{\}\*\s]+|\*)      # tag  
)

| ([^\/\:\[\]\(\)\@\=\{\}\*\s]+|\*)      # tag with no namespace

| (.)             # everything else; used to identify errors
""", re.X).findall
# """""  # fix emacs cruft; having too many special characters fools it
    

def tokenize(s):
    pos = 0
    for token in xpath_tokenizer(s):
        op = token[0]
        if op in ("/", "//"):
            yield (op, None, pos)
        elif token[1]:
            yield ("namespace", (token[1], token[2]), pos)
        elif token[3]:
            yield ("clark", (token[3], token[4]), pos)
        elif token[5]:
            yield ("default", token[5], pos)
        elif token[6]:
            raise SyntaxError("Unknown symbol %r at position %d" %
                              (token[6], pos))
        else:
            raise AssertionError("Unknown token: %r" % (token,))

def _make_original_tag(op, args):
    if op == "namespace":
        return "%s:%s" % (args[0], args[1])
    if op == "clark":
        return "{%s}:%s" % (args[0], args[1])
    if op == "default":
        return args
    raise AssertionError("Strange: %r %r" % (op, args))

def _verify_ordering(tokens):
    if not tokens:
        raise SyntaxError(
            "empty xpath not supported (don't know how to handle that case)")
    pos = 0
    prev = None
    SEP = 1
    URI = 2
    # Check that the path alternates between separator and uri
    for op, args, pos in tokens:
        if op in ("/", "//"):
            if prev == SEP:
                raise SyntaxError(
                   "separator %r may not follow separator at position %d" %
                   (op, pos))
            prev = SEP
        elif op in ("namespace", "clark", "default"):
            if prev == URI:
                errmsg = _make_original_tag(op, args)
                raise SyntaxError(
                   "%r may not follow a separator at position %d" %
                   (errormsg, pos))
            prev = URI
        else:
            raise AssertionError("Unknown op: %r, %r, %r" % (op, args, pos))

    if tokens[-1][0] == "//":
        raise AssertionError("xpath may not end with '//'")

# There are further optimizations.  For example, if this
# returned a match function instead of the regex then it
# could special case terms like /blah//* to mean "startswith('/blah/')"
#   The small performance advantages for most cases doesn't
# currently warrant the extra work.
def to_regexp(s, namespaces={}, default_namespace=None):
    tokens = list(tokenize(s))
    _verify_ordering(tokens)
    
    ### Process the tokens
    re_terms = []
    if tokens[0][0] == "/":
        re_terms.append("^")
        tokens.pop(0)

    for op, args, pos in tokens:
        if op == "/":
            pass
        elif op == "//":
            re_terms.append("(/[^/]+)*")
        elif op in ("namespace", "clark", "default"):
            # Break each apart to get the correct namespace and tag
            if op == "namespace":
                namespace, tag = args
                try:
                    full_namespace = namespaces[namespace]
                except KeyError:
                    raise SyntaxError("Unknown namespace %r at position %d" %
                                      (namespace, pos))
            elif op == "clark":
                full_namespace, tag = args
            elif op == "default":
                full_namespace = default_namespace
                tag = args

            # Figure out which pattern to use for the combination
            # of (namespace, namespace==None) x (tag, tag=='*')
            if full_namespace is None:
                # No namespace specified
                if tag == "*":
                    # Select everything between the /s
                    re_terms.append("/[^/]+")
                else:
                    # Select exactly the tag, no namespace
                    re_terms.append("/%s" % (re.escape(tag),))
            else:
                # namespace specified
                if tag == "*":
                    # Select only fields in the given namespace
                    re_terms.append("/" +
                                    re.escape("{%s}" % (full_namespace,)) +
                                    "[^/]+")
                else:
                    # Must match namespace and tag, exactly
                    re_terms.append("/" +
                                    re.escape("{%s}%s" % (full_namespace, tag)))
        else:
            raise AssertionError("Unknown op %r" % (op,))

    # Must be a complete match
    re_terms.append("/$")
    
    return "".join(re_terms)

class IterParseFilter(object):
    def __init__(self, namespaces=None, default_namespace=None, validate_dtd=False):
        if namespaces is None:
            namespaces = {}
        self.namespaces = namespaces
        self.default_namespace = default_namespace
        self.validate_dtd = validate_dtd

        self._start_document_handlers = []
        self._end_document_handlers = []
        
        self._start_filters = []
        self._end_filters = []
        self._default_start_filters = []
        self._default_end_filters = []
        self._iter_start_filters = []
        self._iter_end_filters = []

        self._start_ns_handlers = []
        self._end_ns_handlers = []
        self._iter_start_ns = False
        self._iter_end_ns = False

    def on_start_document(self, handler):
        self._start_document_handlers.append(handler)
    def on_end_document(self, handler):
        self._end_document_handlers.append(handler)
        
    def _add_handler(self, filters, path, handler):
        path_re = to_regexp(path,
                            namespaces = self.namespaces,
                            default_namespace = self.default_namespace)
        filters.append( (path, re.compile(path_re).search, handler) )
    def on_start(self, path, handler):
        self._add_handler(self._start_filters, path, handler)
    def on_end(self, path, handler):
        self._add_handler(self._end_filters, path, handler)
    def on_start_default(self, path, handler):
        self._add_handler(self._default_start_filters, path, handler)
    def on_end_default(self, path, handler):
        self._add_handler(self._default_end_filters, path, handler)

    def _add_yielder(self, yielders, path):
        path_re = to_regexp(path,
                            namespaces = self.namespaces,
                            default_namespace = self.default_namespace)
                            
        yielders.append( (path, re.compile(path_re).search) )
    def iter_start(self, path):
        self._add_yielder(self._iter_start_filters, path)
    def iter_end(self, path):
        self._add_yielder(self._iter_end_filters, path)

    def on_start_ns(self, handler):
        self._start_ns_handlers.append(handler)
    def on_end_ns(self, handler):
        self._end_ns_handlers.append(handler)
    def iter_start_ns(self):
        self._iter_start_ns = True
    def iter_end_ns(self):
        self._iter_end_ns = True

    def _get_filter_info(self, category):
        for (_, _, pat, handler) in self.filters[category]:
            yield (pat, handler)

    def create_fa(self):
        # Make copies of everything to emphasize that they must
        # not be changed during processing.
        return FilterAutomata(
            start_document_handlers = self._start_document_handlers,
            end_document_handlers = self._end_document_handlers[::-1], # reverse!
            start_filters = self._start_filters[:],
            end_filters = self._end_filters[::-1],  # reversing here!
            default_start_filters = self._default_start_filters[:],
            default_end_filters = self._default_end_filters[::-1], # reversing!
            iter_start_filters = self._iter_start_filters[:],
            iter_end_filters = self._iter_end_filters[:],

            start_ns_handlers = self._start_ns_handlers[:],
            end_ns_handlers = self._end_ns_handlers[::-1],  # reversing here!
            iter_start_ns = self._iter_start_ns,
            iter_end_ns = self._iter_end_ns)

    # These forward to the underlying automata; make a new one each time.
    def parse(self, file, state=None):
        return self.create_fa().parse(file, state, self.validate_dtd)

    # Experimental
    def iterparse(self, file):
        return self.create_fa().iterparse(file, self.validate_dtd)
    # I need a better name
    def handler_parse(self, file, state=None):
        return self.create_fa().handler_parse(file, state)
            

class FilterAutomata(object):
    def __init__(self,
                 start_document_handlers,
                 end_document_handlers,
                 
                 start_filters,
                 end_filters,
                 default_start_filters,
                 default_end_filters,
                 iter_start_filters,
                 iter_end_filters,

                 start_ns_handlers,
                 end_ns_handlers,
                 iter_start_ns,
                 iter_end_ns):
        self.start_document_handlers = start_document_handlers
        self.end_document_handlers = end_document_handlers
        
        self.start_filters = start_filters
        self.end_filters = end_filters
        self.default_start_filters = default_start_filters
        self.default_end_filters = default_end_filters
        self.iter_start_filters = iter_start_filters
        self.iter_end_filters = iter_end_filters
        
        self.start_ns_handlers =  start_ns_handlers
        self.end_ns_handlers = end_ns_handlers
        self.iter_start_ns = iter_start_ns
        self.iter_end_ns = iter_end_ns

        # Can cache results over multiple invocations
        # NOTE: not thread-safe.  Though given the GIL
        # this shouldn't be a problem.
        self.dfa = {}

    def _new_node(self, stack_as_path):
        start_handlers = []
        for (path, matcher, handler) in self.start_filters:
            if matcher(stack_as_path):
                start_handlers.append(handler)
                
        if not start_handlers:
            # Any defaults?
            for (path, matcher, handler) in self.default_start_filters:
                if matcher(stack_as_path):
                    start_handlers.append(handler)

        end_handlers = []
        for (path, matcher, handler) in self.end_filters:
            if matcher(stack_as_path):
                end_handlers.append(handler)
        if not end_handlers:
            # Any defaults?
            for (path, matcher, handler) in self.default_end_filters:
                if matcher(stack_as_path):
                    end_handlers.append(handler)
                            
        # Have all the handlers, now check for yields
        iter_start = False
        for (path, matcher) in self.iter_start_filters:
            if matcher(stack_as_path):
                iter_start = True
                break
            
        iter_end = False
        for (path, matcher) in self.iter_end_filters:
            if matcher(stack_as_path):
                iter_end = True
                break

        new_node = ({}, start_handlers, end_handlers, iter_start, iter_end)
        return new_node

    def _needed_actions(self, iter=False, handler=False):
        if (not handler) and (not cb):
            raise AssertionError("must specify one")
        actions = ("start", "end")
        if ( (handler and self.start_ns_handlers) or
             (iter and self.iter_start_ns) ):
            actions = actions + ("start-ns",)

        if ( (handler and self.end_ns_handlers) or
             (iter and self.iter_end_ns) ):
            actions = actions + ("end-ns",)
        return actions

    # I plan to implement 'handler_parse' as a near copy of 'parse'
    # but without any yield statements.
    def handler_parse(self, file, state=None):
        for x in self.parse(file, state):
            pass
        
    # I plan to implement 'iterparse' as a near copy of 'parse'
    # but without any references to callbacks
    def iterparse(self, file, validate_dtd=False):
        return self.parse(file, None, validate_dtd)

    def parse(self, file, state=None, validate_dtd=False):
        if not dtd_validation:
            validate_dtd = False
        node_stack = []
        node_stack_append = node_stack.append
        tag_stack = []
        tag_stack_append = tag_stack.append
        # children, start handlers, end handlers, iter start, iter end
        node = (self.dfa, [], [], False, False)

        # synthesize start-document events
        for handler in self.start_document_handlers:
            handler("start-document", None, state)

        # figure out if I also need start-ns and/or end-ns events
        needed_actions = self._needed_actions(True, True)
        kwargs = {}
        if validate_dtd:
            kwargs = dict(dtd_validation=True)
        last_start = 0
        total_mem = 0
        before = None
        for (event, ele) in etree.iterparse(file, needed_actions, **kwargs):
            if event == "start":
                tag = ele.tag
                # Descend into node; track where I am
                tag_stack_append(tag)
                node_stack_append(node)
                stack_as_path = "/" + ("/".join(tag_stack)) + "/"
                new_node = self._new_node(stack_as_path)
                node = new_node

                # call the start handlers then yield the element
                for start_handler in node[1]:
                    start_handler(event, ele, state)
                if node[3]:
                    yield (event, ele)
                #print total_mem

            elif event == "end":
                # call the end handlers then yield the element
                for end_handler in node[2]:
                    end_handler(event, ele, state)
                del tag_stack[-1]
                if node[4]:
                    yield (event, ele)
                    # It's safe to call clear() here because no descendants will be
                    # accessed
                    ele.clear()
                    ele.getparent().remove(ele)

                    # Also eliminate now-empty references from the root node to elem
                    #for ancestor in ele.xpath('ancestor-or-self::*'):
                    #    while ancestor.getprevious() is not None:
                    #        del ancestor.getparent()[0]
                node = node_stack.pop()

            elif event == "start-ns":
                for handler in self.start_ns_handlers:
                    handler(event, ele, state)
                if self.iter_start_ns:
                    print('start-ns')
                    yield (event, ele)
                
            elif event == "end-ns":
                for handler in self.end_ns_handlers:
                    handler(event, ele, state)
                if self.iter_start_ns:
                    print('end-ns')
                    yield (event, ele)
                    # It's safe to call clear() here because no descendants will be
                    # accessed
                    ele.clear()
                    ele.getparent().remove(ele)
                    # Also eliminate now-empty references from the root node to elem
                    #for ancestor in ele.xpath('ancestor-or-self::*'):
                    #    while ancestor.getprevious() is not None:
                    #        del ancestor.getparent()[0]
                    
        for handler in self.end_document_handlers:
            handler("end-document", None, state)


#### An incomplete test suite ####

def test_path(path, args):
    #print "**** test_path", repr(path), repr(args)
    pattern = to_regexp(path)
    pat = re.compile(pattern)
    s = "/" + ("/".join(args)) + "/"
    #print pattern, s
    return bool(pat.search(s))

def test_ns_path(path, args):
    #print "**** test_path", repr(path), repr(args)
    pattern = to_regexp(path,
                        namespaces = {
        "xml": "http://www.w3.org/XML/1998/namespace",
        "das2": "http://biodas.org/documents/das2"},
                        # the empty namespace is not the same as no namespace!
                        default_namespace = "")
        
    pat = re.compile(pattern)
    s = "/" + ("/".join(args)) + "/"
    #print pattern, s
    return bool(pat.search(s))

def test_syntax():
    for (xpath, tag_list, expect) in (
           ("A", ["A"], 1),
           ("A", ["AA"], 0),
           ("A", ["B", "A"], 1),
           ("/A", ["B", "A"], 0),
           ("/B", ["B", "A"], 0),
           ("//A", ["B", "A"], 1),
           ("A//B", ["A", "B"], 1),
           ("A//B", ["C", "A", "B"], 1),
           ("/A//B", ["C", "A", "B"], 0),
           ("/B/*", ["B", "A"], 1),
           # Test back-tracking; both greedy and non-greedy cases
           ("A//B//C//D", ["A", "B", "C", "B", "D"], 1),
           ("A//B/D", ["A", "B", "C", "B", "D"], 1),

           # Clark namespace tests
           ("{http://x.com}A", ["{http://x.com}A"], 1),
           ("{http://x.org}A", ["{http://x.com}A"], 0),
           ("{http://x.org}A", ["{http://x.com}B", "{http://x.org}A"], 1),
           ("*", ["{http://x.com}A"], 1),
           ("{http://x.com}*", ["{http://x.com}A"], 1),
           ("{http://x.com}*", ["{http://x.org}A"], 0),
           
           ):
        got = test_path(xpath, tag_list)
        if got != expect:
            raise AssertionError("xpath %r against %r got %r, expected %r" %
                                 (xpath, tag_list, got, bool(expect)))

    for (xpath, tag_list, expect) in (
           # various namespace checks
           ("xml:A", ["{http://www.w3.org/XML/1998/namespace}A"], 1),
           ("xml:A", ["{http://www.w3.org/XML/1998/namespace2}A"], 0),
           ("xml:A", ["{http://www.w3.org/XML/1998/namespace}AA"], 0),
           ("xml:A", ["{http://www.w3.org/XML/1998/namespace}B",
                      "{http://www.w3.org/XML/1998/namespace}A"], 1),
           ("xml:B", ["{http://www.w3.org/XML/1998/namespace}B",
                      "{http://www.w3.org/XML/1998/namespace}A"], 0),

           ("A", ["{}A"], 1),
           ("A", ["A"], 0),

           ("*", ["A"], 0),
           ("*", ["{}A"], 1),
           ("das2:*", ["{http://biodas.org/documents/das2}AAA"], 1),
           ("das2:*", ["{}AAA"], 0),
           ("xml:*/das2:*", ["{http://www.w3.org/XML/1998/namespace}ABC",
                             "{http://biodas.org/documents/das2}ABC"], 1),
           ("das2:*/xml:*", ["{http://www.w3.org/XML/1998/namespace}ABC",
                             "{http://biodas.org/documents/das2}ABC"], 0),
           

           ):
        got = test_ns_path(xpath, tag_list)
        if got != expect:
            raise AssertionError("xpath %r against %r got %r, expected %r" %
                                 (xpath, tag_list, got, bool(expect)))

def test_filtering():
    import cStringIO as StringIO
    f = StringIO.StringIO("""\
 <A><AA>
   <B xmlns="http://z/"><C/><spam:D xmlns:spam="http://spam/">eggs</spam:D></B>
   <B x='6'>foo<B y='7'>bar</B>baz</B>
 </AA></A>""")
    special = object()
    class Capture(object):
        def __init__(self):
            self.history = []
        def __call__(self, event, ele, state):
            if state is not special:
                raise AssertionError("Did not get expected state")
            self.history.append( (event, ele) )

    filter = IterParseFilter()
    capture_all = Capture()
    filter.on_start_document(capture_all)
    filter.on_start("*", capture_all)
    filter.on_end("*", capture_all)
    filter.on_end_document(capture_all)
    filter.on_start_ns(capture_all)
    filter.on_end_ns(capture_all)

    for x in filter.parse(f, state=special):
        raise AssertionError("should not yield %r" % (x,))

    expect_history = (
        ("start-document", None),
        ("start", "A"),
        ("start", "AA"),
        ("start-ns", ("", "http://z/")),
        ("start", "{http://z/}B"),
        ("start", "{http://z/}C"),
        ("end", "{http://z/}C"),
        ("start-ns", ("spam", "http://spam/")),
        ("start", "{http://spam/}D"),
        ("end", "{http://spam/}D"),
        ("end-ns", None),
        ("end", "{http://z/}B"),
        ("end-ns", None),
        ("start", "B"),
        ("start", "B"),
        ("end", "B"),
        ("end", "B"),
        ("end", "AA"),
        ("end","A"),
        ("end-document", None),
        )

    for (got, expect) in zip(capture_all.history, expect_history):
        event, ele = got
        tag = getattr(ele, "tag", ele)
        if (event, tag) != expect:
            raise AssertionError("Expected %r Got %r" % (expect, (event, tag)))
    if len(capture_all.history) != len(expect_history):
        raise AssertionError("Length mismatch")
            
    f.seek(0)
    filter = IterParseFilter()
    def must_match_B(event, ele, state):
        if ele.tag != "B":
            raise AssertionError("%r is not B" % (ele.tag,))
    def must_match_B_y7(event, ele, state):
        if ele.tag != "B":
            raise AssertionError("%r is not B" % (ele.tag,))
        if ele.attrib["y"] != "7":
            raise AssertionError("%r is not the correct B" % (ele.tag,))

    filter.on_start("B", must_match_B)
    filter.on_start("B/B", must_match_B_y7)
    
    f.seek


def test_parse():
    import os
    filename = "/Users/dalke/Music/iTunes/iTunes Music Library.xml"
    if not os.path.exists(filename):
        print ("Cannot find %r: skipping test" % (filename,))
        return

    # Work through callbacks
    ef = IterParseFilter()
    def print_info(event, ele, state):
        d = {}
        children = iter(ele)
        for child in children:
            key = child.text
            value = children.next().text
            d[key] = value
        print ("%r is by %r" % (d["Name"], d.get("Artist", "<unknown>")))
        ele.clear()
            
    ef.on_end("/plist/dict/dict/dict", print_info)
    ef.handler_parse(open(filename))

    # Work through iterators
    ef = IterParseFilter()
    ef.iter_end("/plist/dict/dict/dict")
    for (event, ele) in ef.iterparse(open(filename)):
        d = {}
        children = iter(ele)
        for child in children:
            key = child.text
            value = children.next().text
            d[key] = value
        print ("%r is a %r song" % (d["Name"], d.get("Genre", "<unknown>")))
        ele.clear()
            

def test():
    test_syntax()
    test_filtering()
    test_parse()

if __name__ == "__main__":
    test()
    print ("All tests passed.")
